#' Data
#'
#' Data encapsulation
#'
#' @docType class
#' @importFrom R6 R6Class
#'
#' @field .data List of
#' \itemize{
#'   \item time: vector of times
#'   \item centered_serie: centered series
#'   \item level: corresponding levels
#'   \item exo: exogenous variables
#'   \item exo_hat: predicted exogenous variables
#' }
#'
#' @section Methods:
#' \describe{
#' \item{\code{getSize()}}{
#'   Return number of series in dataset.}
#' \item{\code{getStdHorizon()}}{
#'   Return number of time steps from serie[1] until midnight}
#' \item{\code{appendHat(time, exo_hat)}}{
#'   New estimated exogenous variables + time}
#' \item{\code{append(serie, exo)}}{
#'   New measured data; call *after* \code{appendHat()}}
#' \item{\code{getTime(index)}}{
#'   Get times at specified index.}
#' \item{\code{getCenteredSerie(index)}}{
#'   Get (measured or predicted) centered serie at specified index.}
#' \item{\code{getCenteredSeries(indices)}}{
#'   Get centered series at specified indices (in columns).}
#' \item{\code{getLevel(index)}}{
#'   Get level at specified index.}
#' \item{\code{getSerie(index)}}{
#'   Get serie (centered+level) at specified index.}
#' \item{\code{getSeries(indices)}}{
#'   Get series at specified indices (in columns).}
#' \item{\code{getExoHat(index)}}{
#'   Get predicted exogenous variables at specified index.}
#' \item{\code{getExo(index)}}{
#'   Get exogenous variables at specified index.}
#' \item{\code{removeFirst()}}{
#'   Remove first list element (if truncated).}
#' \item{\code{removeLast()}}{
#'   Remove last list element (if truncated).}
#' }
#'
Data = R6::R6Class("Data",
	private = list(
		.data = list()
	),
	public = list(
		getSize = function()
			length(private$.data)
		,
		getStdHorizon = function()
			24 - as.POSIXlt( private$.data[[1]]$time[1] )$hour + 1
		,
		append = function(time, serie, exo, exo_hat)
		{
			level = mean(serie, na.rm=TRUE)
			centered_serie = serie - level
			private$.data[[length(private$.data)+1]] <- list(
				"time"=time, #H-24 --> H-1
				"centered_serie"=centered_serie, #at 'time'
				"level"=level, #at 'time'
				"exo"=exo, #at 'time' (yersteday 0am to last midnight)
				"exo_hat"=exo_hat) #today 0am to next midnight
		},
		getTime = function(index)
		{
			index = dateIndexToInteger(index, self)
			private$.data[[index]]$time
		},
		getCenteredSerie = function(index)
		{
			index = dateIndexToInteger(index, self)
			private$.data[[index]]$centered_serie
		},
		getCenteredSeries = function(indices)
			sapply(indices, function(i) self$getCenteredSerie(i))
		,
		getLevel = function(index)
		{
			index = dateIndexToInteger(index, self)
			private$.data[[index]]$level
		},
		getSerie = function(index)
		{
			index = dateIndexToInteger(index, self)
			private$.data[[index]]$centered_serie + private$.data[[index]]$level
		},
		getSeries = function(indices)
			sapply(indices, function(i) self$getSerie(i))
		,
		getExoHat = function(index)
		{
			index = dateIndexToInteger(index, self)
			private$.data[[index]]$exo_hat
		},
		getExo = function(index)
		{
			index = dateIndexToInteger(index, self)
			private$.data[[index]]$exo
		},
		removeFirst = function()
			private$.data <- private$.data[2:length(private$.data)]
		,
		removeLast = function()
			private$.data <- private$.data[1:(length(private$.data)-1)]
	)
)
