#' Neighbors Forecaster
#'
#' Predict tomorrow as a weighted combination of "futures of the past" days.
#' Inherits \code{\link{Forecaster}}
#'
NeighborsForecaster = R6::R6Class("NeighborsForecaster",
	inherit = Forecaster,

	public = list(
		predictShape = function(data, today, memory, horizon, ...)
		{
			# (re)initialize computed parameters
			private$.params <- list("weights"=NA, "indices"=NA, "window"=NA)

			# Do not forecast on days with NAs (TODO: softer condition...)
			if (any(is.na(data$getCenteredSerie(today))))
				return (NA)

			# Determine indices of no-NAs days followed by no-NAs tomorrows
			fdays = getNoNA2(data, max(today-memory,1), today-1)

			# Get optional args
			local = ifelse(hasArg("local"), list(...)$local, TRUE) #same level + season?
			simtype = ifelse(hasArg("simtype"), list(...)$simtype, "none") #or "endo", or "exo"
			if (hasArg("window"))
			{
				return ( private$.predictShapeAux(data,
					fdays, today, horizon, local, list(...)$window, simtype, TRUE) )
			}

			# Indices of similar days for cross-validation; TODO: 20 = magic number
			cv_days = getSimilarDaysIndices(today, data, limit=20, same_season=FALSE,
				days_in=fdays)

			# Optimize h : h |--> sum of prediction errors on last N "similar" days
			errorOnLastNdays = function(window, simtype)
			{
				error = 0
				nb_jours = 0
				for (i in seq_along(cv_days))
				{
					# mix_strategy is never used here (simtype != "mix"), therefore left blank
					prediction = private$.predictShapeAux(data, fdays, cv_days[i], horizon, local,
						window, simtype, FALSE)
					if (!is.na(prediction[1]))
					{
						nb_jours = nb_jours + 1
						error = error +
							mean((data$getSerie(cv_days[i]+1)[1:horizon] - prediction)^2)
					}
				}
				return (error / nb_jours)
			}

			# TODO: 7 == magic number
			if (simtype=="endo" || simtype=="mix")
			{
				best_window_endo = optimize(
					errorOnLastNdays, c(0,7), simtype="endo")$minimum
			}
			if (simtype=="exo" || simtype=="mix")
			{
				best_window_exo = optimize(
					errorOnLastNdays, c(0,7), simtype="exo")$minimum
			}

			best_window =
				if (simtype == "endo")
					best_window_endo
				else if (simtype == "exo")
					best_window_exo
				else if (simtype == "mix")
					c(best_window_endo,best_window_exo)
				else #none: value doesn't matter
					1

			return(private$.predictShapeAux(data, fdays, today, horizon, local,
				best_window, simtype, TRUE))
		}
	),
	private = list(
		# Precondition: "today" is full (no NAs)
		.predictShapeAux = function(data, fdays, today, horizon, local, window, simtype,
			final_call)
		{
			fdays_cut = fdays[ fdays < today ]
			if (length(fdays_cut) <= 1)
				return (NA)

			if (local)
			{
				# Neighbors: days in "same season"; TODO: 60 == magic number...
				fdays = getSimilarDaysIndices(today, data, limit=60, same_season=TRUE,
					days_in=fdays_cut)
				if (length(fdays) <= 1)
					return (NA)
				levelToday = data$getLevel(today)
				distances = sapply(fdays, function(i) abs(data$getLevel(i)-levelToday))
				#TODO: 2, 10, 3, 12 magic numbers here...
				dist_thresh = 2
				min_neighbs = min(10,length(fdays))
				repeat
				{
					same_pollution = (distances <= dist_thresh)
					nb_neighbs = sum(same_pollution)
					if (nb_neighbs >= min_neighbs) #will eventually happen
						break
					dist_thresh = dist_thresh + 3
				}
				fdays = fdays[same_pollution]
				max_neighbs = 12
				if (nb_neighbs > max_neighbs)
				{
					# Keep only max_neighbs closest neighbors
					fdays = fdays[
						sort(distances[same_pollution],index.return=TRUE)$ix[1:max_neighbs] ]
				}
				if (length(fdays) == 1) #the other extreme...
				{
					if (final_call)
					{
						private$.params$weights <- 1
						private$.params$indices <- fdays
						private$.params$window <- 1
					}
					return ( data$getSerie(fdays[1])[1:horizon] )
				}
			}
			else
				fdays = fdays_cut #no conditioning

			if (simtype == "endo" || simtype == "mix")
			{
				# Compute endogen similarities using given window
				window_endo = ifelse(simtype=="mix", window[1], window)

				# Distances from last observed day to days in the past
				serieToday = data$getSerie(today)
				distances2 = sapply(fdays, function(i) {
					delta = serieToday - data$getSerie(i)
					mean(delta^2)
				})

				sd_dist = sd(distances2)
				if (sd_dist < .25 * sqrt(.Machine$double.eps))
				{
#					warning("All computed distances are very close: stdev too small")
					sd_dist = 1 #mostly for tests... FIXME:
				}
				simils_endo = exp(-distances2/(sd_dist*window_endo^2))
			}

			if (simtype == "exo" || simtype == "mix")
			{
				# Compute exogen similarities using given window
				window_exo = ifelse(simtype=="mix", window[2], window)

				M = matrix( nrow=1+length(fdays), ncol=1+length(data$getExo(today)) )
				M[1,] = c( data$getLevel(today), as.double(data$getExo(today)) )
				for (i in seq_along(fdays))
					M[i+1,] = c( data$getLevel(fdays[i]), as.double(data$getExo(fdays[i])) )

				sigma = cov(M) #NOTE: robust covariance is way too slow
				# TODO: 10 == magic number; more robust way == det, or always ginv()
				sigma_inv =
					if (length(fdays) > 10)
						solve(sigma)
					else
						MASS::ginv(sigma)

				# Distances from last observed day to days in the past
				distances2 = sapply(seq_along(fdays), function(i) {
					delta = M[1,] - M[i+1,]
					delta %*% sigma_inv %*% delta
				})

				sd_dist = sd(distances2)
				if (sd_dist < .25 * sqrt(.Machine$double.eps))
				{
#					warning("All computed distances are very close: stdev too small")
					sd_dist = 1 #mostly for tests... FIXME:
				}
				simils_exo = exp(-distances2/(sd_dist*window_exo^2))
			}

			similarities =
				if (simtype == "exo")
					simils_exo
				else if (simtype == "endo")
					simils_endo
				else if (simtype == "mix")
					simils_endo * simils_exo
				else #none
					rep(1, length(fdays))
			similarities = similarities / sum(similarities)

			prediction = rep(0, horizon)
			for (i in seq_along(fdays))
				prediction = prediction + similarities[i] * data$getSerie(fdays[i]+1)[1:horizon]

			if (final_call)
			{
				private$.params$weights <- similarities
				private$.params$indices <- fdays
				private$.params$window <-
					if (simtype=="endo")
						window_endo
					else if (simtype=="exo")
						window_exo
					else if (simtype=="mix")
						c(window_endo,window_exo)
					else #none
						1
			}

			return (prediction)
		}
	)
)
